from typing import List
from ftplib import error_perm
from io import StringIO
import json
import os
import shlex
import subprocess
import zipfile

def decode_json(command) -> json:
    """[summary]

        Return Azure data in json decoded version
        
    Args:
        command (str): Command argument given

    Returns:
        [json]: JSON encoded
    """
    # Converto command to popen format
    args = shlex.split(command)
    
    ## Get app service list
    #  Open process
    p = subprocess.Popen(args, stdout=subprocess.PIPE)
    
    # Create IO data reader
    io = StringIO(p.stdout.read().decode())
    
    # Formato io data reader in JSON
    return json.load(io)

def unzipFiles(path, zip) -> List:
    """[summary]

        Extract zip file on defined path and return exctract file location

    Args:
        path (str): Path where zip file will be stored
    """
    if zip.endswith('.zip'):
        filePath=zip
        zip_file = zipfile.ZipFile(filePath)
        for names in zip_file.namelist():
            zip_file.extract(names,path)
        zip_file.close() 
        return zip_file.namelist()[0]


def listZipFiles(path, zip) -> List:
    """[summary]

        List zip file on defined path and return exctract file location

    Args:
        path (str): Path where zip file will be stored
    """
    if zip.endswith('.zip'):
        filelist = []
        filePath=zip
        zip_file = zipfile.ZipFile(filePath)
        for names in zip_file.namelist():
            filelist.append(path+names)
        zip_file.close() 
        return filelist


from ftplib import FTP, FTP_TLS


def init_ftp(url, ssl=False) -> FTP:
    """[summary]

    Generate FTP Connection
    
    Args:
        url (str): Url to connect

    Returns:
        FTP: FTP Connection
    """
    url = url.replace("ftp://", "")
    url = url.replace("/site/wwwroot", "")
    
    if ssl:
        try:
            return FTP_TLS(url)
        except Exception as error:
            raise("Cannot estabilish TLS connection: {0}"
                  .format(error))
    else:
        try:
            return FTP(url)
        except Exception as error:
            raise("Cannot estabilish connection: {0}"
                  .format(error))


from ftplib import error_perm
import sys
from appsrvdeployer.modules.ftp import init_ftp
from appsrvdeployer.modules.logger import *
from appsrvdeployer.modules.utils import decode_json, unzipFiles
import os


def uploadFiles(ftp, path, logger, location=None):
    """[summary]

        Upload exctracted file on app services
    Args:
        ftp (FTP): Ftp connection
        path (str): Extracted file zip path
    """
    
    # Change ftp dir
    if location:
        ftp.cwd(location) 
    
    for name in os.listdir(path):
        localpath = os.path.join(path, name)
        if os.path.isfile(localpath):
            logger.info("STOR %s %s" % (name, localpath))
            ftp.storbinary('STOR ' + name, open(localpath,'rb'))
        elif os.path.isdir(localpath):
            logger.info("MKD %s" % (name))

            try:
                ftp.mkd(name)

            # ignore "directory already exists"
            except error_perm as e:
                if not e.args[0].startswith('550'): 
                    raise

            logger.info("CWD %s" % (name))
            ftp.cwd(name)
            uploadFiles(ftp, localpath, logger)           
            logger.debug("CWD ..")
            ftp.cwd("..")
            

def provisioning(
        j_appsrv, 
        rg, 
        subscription, 
        path, 
        zip,
        dirpath,
        logfile
    ):
    
    for app in j_appsrv:
        
        # Create App Logger 
        logger = create_logger(
            app_name=app,
            log_level=os.environ.get("APPSRVDEPLOYER_LOG_LEVEL") 
                if os.environ.get("APPSRVDEPLOYER_LOG_LEVEL") 
                else logging.INFO,
            logfile=logfile,
            stdout=True,
            file=True
        )
        
        logger.info("Retrive information from {0}...".format(app))
        
        url = decode_json(
                'az webapp deployment list-publishing-profiles {rg} {sub} --name {0} --query "[1].publishUrl"'
                .format(app, rg=rg, sub=subscription)
            )
        user = decode_json(
                'az webapp deployment list-publishing-profiles {rg} {sub} --name {0} --query "[1].userName"'
                .format(app, rg=rg, sub=subscription)
            )
        passwd = decode_json(
                'az webapp deployment list-publishing-profiles {rg} {sub} --name {0} --query "[1].userPWD"'
                .format(app, rg=rg, sub=subscription)
            )
        
        logger.debug("App service [{app}] connection url [{conn}] : user: {user} pass: {passwd}".format(
                app=app,
                conn=url,
                user=user,
                passwd=passwd,
            )
        )
    
        try:    
            logger.info("Try to create ftp connection to -> {0}".format(url))
            conn = init_ftp(url, ssl=True)
        except Exception as e:
            logger.error("Error while creating connection to -> {0}: {e}".format(url, e=e))
            
        conn.login(user=user, passwd=passwd)
        
        # Case selector
        if path:
            try:
                conn.retrlines("LIST %s" % (path))
            except error_perm as e_perm:
                logger.error("Failed to retrive path on app service, please select anthor destination: {e}"
                    .format(e=e_perm))
                        
                choice = input("Want crete this path? -> {0} ( 'Yes' or 'No') $ "
                    .format(path))

                if choice.lower() == "y" or choice.lower() == "yes":
                    conn.mkd(path)
                else:
                    sys.exit(550)
                    
        if zip:
            try:
                exctdir = unzipFiles(dirpath.name, zip)
            except Exception as e:
                logger.error("Error while unzip files: {e}".format(e=e))
                
            try: 
                uploadFiles(conn, dirpath.name + "/" + exctdir, logger, location=path)
            except Exception as e:
                logger.error("Error while upload files: {e}".format(e=e))
            
        conn.close()


import argparse
from cmath import log
from email.policy import default
import sys
from tempfile import TemporaryDirectory
import os
from datetime import datetime
from appsrvdeployer.modules.logger import *
from appsrvdeployer.modules.utils import listZipFiles, decode_json
from appsrvdeployer.modules.ftp import *
import textwrap
from appsrvdeployer.modules.provisioning import provisioning
from tempfile import mkstemp
import atexit

def exit_handler(logfile):
    print("Can read log file at: {0}".format(logfile))


def init_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=textwrap.dedent("""
                                    Provisioning config file to app service.                                    
                                    Take zip file and upload on app service on defined path.
                                """),
                                formatter_class=argparse.RawDescriptionHelpFormatter,
                                prog="appsrvdeployer",
                                epilog=textwrap.dedent('''\
                                It's recommended to make dry run before run provisioning 
                                to see what\'s app are selected for deployments\n
                                '''))

    azure = parser.add_argument_group("azure options")
    azure.add_argument("-n", "--app-service-name",
                       dest="appsrv_name", default="",
                       help="App service name or like, to deploy into filtered app services")    
    azure.add_argument("--resource-group", '-g', 
                       dest="group",
                       default="")
    azure.add_argument("--subscription", '-s', 
                       dest="subscription", 
                       default="", 
                       required=False)
    
    provision = parser.add_argument_group("provisioning options")
    provision.add_argument("--zip", "-z", 
                           dest="zip",
                           required=True)    
    provision.add_argument("--path", "-p", 
                           dest="path", 
                           type=str,
                           required=True)
    provision.add_argument("--dry-run", "-C", 
                           dest="DRY_RUN",
                           default=False,
                           action="store_true")
    
    utils = parser.add_argument_group("utils")
    utils.add_argument(
        "--log",
        dest="logfile",
        type=str,
        default=mkstemp(prefix="logger-appsrvdeployer-", suffix=".log")[1],
        help="Set log file location."
    )

    return parser.parse_args()

def main() -> None:
    """[summary]
    Main function
    """

    # Init parser arguments
    args = init_parser() 
    
    # Register exit handler
    atexit.register(exit_handler, args.logfile)

    # Create App Logger 
    logger = create_logger(
        app_name="appsrvdeployer",
        log_level=os.environ.get("APPSRVDEPLOYER_LOG_LEVEL") 
            if os.environ.get("APPSRVDEPLOYER_LOG_LEVEL") 
            else logging.INFO,
        logfile=args.logfile,
        stdout=True,
        file=True
    )
    
    # Parsing subscription
    if args.subscription != "":
        args.subscription = "--subscription "+ args.subscription
        
    # Parsing resource group
    if args.group != "":
        args.group = "--resource-group "+ args.group 
    
    # Init temporary dir to storage zip file
    dirpath = TemporaryDirectory(
            prefix="appsrvdeployer-", 
            suffix=datetime.today().strftime('-%Y%m%d%H%M')
        ) 
    
    # Set command to execute
    if args.appsrv_name:
        j = decode_json('az appservice plan list --query "[?contains(name, \'{0}\')].[name]" \
                        {rg} {sub}'
                        .format(args.appsrv_name,
                                rg=args.group, 
                                sub=args.subscription))
        if len(j) == 1:
            j = j[0]
    else:
        j = decode_json("az appservice plan list --query [].name \
                        {rg} {sub}"
                        .format(rg=args.group, 
                                sub=args.subscription))
            

    if args.appsrv_name:
        j_appsrv = decode_json('az webapp list --query "[?contains(name, \'{0}\')].[name]" \
                                {rg} {sub}'
                               .format(args.appsrv_name,
                                       rg=args.group, 
                                       sub=args.subscription))  
        if len(j_appsrv) == 1:
            j_appsrv = j_appsrv[0] 
    else:
        j_appsrv = decode_json('az webapp list --query "[].name" \
            {rg} {sub}'
            .format(rg=args.group, 
                    sub=args.subscription))

    
    # Print app service list 
    logger.info("App service plan list")
    for plan in j:
        logger.info("Found plan: {0}".format(plan))

    if args.DRY_RUN:
        logger.info("Will deploy in {0}"
                    .format(j_appsrv))

        for f in listZipFiles(dirpath.name, args.zip):
            logger.info("Will upload this: {0}".format(f))

        sys.exit(0)
    
    provisioning(
        j_appsrv,
        args.group,
        args.subscription,
        args.path,
        args.zip,
        dirpath,
        args.logfile
    )

if __name__ == "__main__":
    main()


import logging
import logging.handlers
from tempfile import mkstemp

def create_logger(
    app_name, 
    log_level=logging.DEBUG, 
    logfile=mkstemp(prefix="logger-", suffix=".log"), 
    address="", 
    stdout=True, 
    syslog=False, 
    file=False):
    """
    create logging object with logging to syslog, file and stdout
    :param logfile log file
    :param app_name app name
    :param log_level logging log level
    :param address address of syslog
    :param stdout log to stdout
    :param syslog log to syslog
    :param file log to file
    :return: logging object
    """

    # Creazione di un logging e il set del livello di log
    logger = logging.getLogger(app_name)
    logger.setLevel(log_level)

    # Imposto il format di default per il logging
    formatter = logging.Formatter('[%(asctime)s][%(name)s][%(threadName)s] %(levelname)s: %(message)s')

    if file:
        # create file logger handler
        fh = logging.FileHandler(logfile)
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    if syslog:
        # create syslog logger handler
        sh = logging.handlers.SysLogHandler(address = (address if address else "localhost",514), facility='local5', socktype=None)
        sh.setLevel(log_level)
        sf = logging.Formatter('%(message)s')
        sh.setFormatter(sf)
        logger.addHandler(sh)

    if stdout:
        # create stream logger handler
        ch = logging.StreamHandler()
        ch.setLevel(log_level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    return logger